import hashlib
from datetime import timedelta

import jwt
from django.conf import settings
from django.http import HttpResponse, HttpResponseNotModified
from django.utils import timezone
from django.utils.http import http_date
from django.utils.module_loading import import_string
from django.views.generic import TemplateView

from django_includes.jinja2 import get_markup


def include_view(request, token, via):
    data = jwt.decode(token, settings.SECRET_KEY)
    view = import_string(data['v'])
    if hasattr(view, 'as_view'):  # xxx not a good idea, better use already constructed views.
        view = view.as_view()

    # build wrapped response
    request.META['HTTP_X_REQUESTED_WITH'] = 'XMLHttpRequest'
    wrapped_response = view(request, *data['a'], **data['k'])
    if wrapped_response.status_code >= 300:
        return wrapped_response

    response = HttpResponse(
        get_markup(view, wrapped_response, via)
    )

    for k, v in wrapped_response.items():
        response[k] = v

    return response


class CacheableTemplateView(TemplateView):
    template_name = 'promised_views_app/cacheable.html'
    use_etag = True
    cache_control = 'private'
    max_age = 86400

    _etag = None

    def get_etag(self):
        if not self.use_etag:
            return None

        if not self._etag:
            # self.etag = hashlib.sha1(repr(args).encode('utf-8') + b';' + repr(sorted(kwargs.items())).encode('utf-8')).hexdigest()
            self._etag = hashlib.sha1(repr(type(self)).encode()).hexdigest()
        return self._etag

    def dispatch(self, request, *args, **kwargs):
        # if we're using etag and the browser knows about the actual page etag, we can bypass dispatching the request
        # for real and just tell the browser its cache is up to date.
        etag = self.get_etag()
        if etag and etag == request.META.get('HTTP_IF_NONE_MATCH', None):
            return HttpResponseNotModified()

        # time to get to work, my dear.
        response = super().dispatch(request, *args, **kwargs)

        # only cache if secondary request
        if request.is_ajax():
            # build expires header. Not used in http 1.1, but maybe useful for older browsers? Or maybe completely
            # useless, but also harmless.
            if self.max_age:
                response['Expires'] = http_date((timezone.now() + timedelta(seconds=self.max_age)).timestamp())

            # build cache-control header
            if self.cache_control:
                cache_control = [self.cache_control]
                if self.max_age and not self.cache_control.startswith('no-'):
                    cache_control.append('max-age={}'.format(self.max_age))
                response['Cache-Control'] = ', '.join(cache_control)

            if etag:
                response['ETag'] = etag

        return response


cacheable_view = CacheableTemplateView.as_view()



# coding: utf-8

from __future__ import absolute_import, print_function, unicode_literals

def patch_http_development_server():
    """
    HACK: without HTTP/1.1, Chrome ignores certain cache headers during development!
     see http://stackoverflow.com/a/28033770/179583 for a bit more discussion.

    You can include this in your wsgi.py file.

    """
    from wsgiref import simple_server
    simple_server.ServerHandler.http_version = "1.1"




import html
import jwt
import six
from django.conf import settings
from django.core.urlresolvers import reverse
from django.utils.module_loading import import_string
from jinja2.ext import Extension
from markupsafe import Markup

from django_includes.settings import DEBUG_INCLUDES


def get_view_class(view_or_callable):
    try:
        return view_or_callable.view_class
    except AttributeError as e:
        return type(view_or_callable)


def get_markup(view, response, via='render'):
    # headers
    headers = '<br>'.join('{} = {}'.format(k, v) for k, v in response.items())
    if headers:
        headers = '<br>' + headers

    # render
    if hasattr(response, 'render'):
        response.render()

    if DEBUG_INCLUDES:
        return Markup(('<div style="outline: 1px solid #B9B280; padding: 1em; position: relative;">'
                       '  <em style="color: white; position: absolute; right: -1px; top: -1px; padding: 0.15em 0.5em; font-size: 10px; background-color: #B9B280">'
                       '    {} ({}){}'
                       '  </em>'
                       '  {} '
                       '</div>').format(
            html.escape(get_view_class(view).__name__),
            via,
            headers,
            response.render().content.decode(encoding='utf-8'),
        ))

    return Markup(response.content.decode(encoding='utf-8'))


def render(request, view, *args, **kwargs):
    if isinstance(view, six.string_types):
        view = import_string(view)

    if hasattr(type(view), 'as_view'):
        view = type(view).as_view()

    response = view(request, *args, **kwargs)

    request.META['HTTP_X_REQUESTED_WITH'] = 'XMLHttpRequest'

    return get_markup(view, response, 'render')


def include_tag(route, request, view_name, view_args, view_kwargs):
    if route == 'hinclude':
        tag = 'hx:include'
    elif route == 'esi':
        tag = 'esi:include'
    else:
        raise ValueError('Unsupported inclusion type.')

    return Markup('<{} src="{}" />'.format(
        tag,
        request.build_absolute_uri(reverse(route, kwargs={
            'token': jwt.encode({
                'v': view_name,
                'a': view_args,
                'k': view_kwargs
            }, settings.SECRET_KEY, )
        }))))


def render_hinclude(request, view_name, *args, **kwargs):
    return include_tag('hinclude', request, view_name, args, kwargs)


def render_esi(request, view_name, *args, **kwargs):
    return include_tag('esi', request, view_name, args, kwargs)


class DjangoIncludesExtension(Extension):
    def __init__(self, environment):
        super(DjangoIncludesExtension, self).__init__(environment)
        environment.globals['render_sync'] = render
        environment.globals['render_hinclude'] = render_hinclude
        environment.globals['render_esi'] = render_esi



# coding: utf-8

from __future__ import absolute_import, print_function, unicode_literals

import re
from http.cookiejar import split_header_words

import cachecontrol
import requests

esi_include_re = re.compile(b'<esi:include src="([^"]+)" />')


class EdgeSideIncludesMiddleware(object):
    """
    This middleware is intended to resolve edge side includes, as would do a system side middleware like Varnish. It is
    by no mean complete, efficient or secure, please consider using a real implementation if running in production.

    Please note that this is a WSGI middleware, and not a django middleware. Use it by patching your wsgi.py file to
    decorate the django application.

    """

    def __init__(self, application, debug=False):
        self.application = application
        self.environ = None
        if debug:
            self.session = requests.session()
        else:
            self.session = cachecontrol.CacheControl(requests.session())

    def include(self, match):
        # cookies forwarding is required, especially for session and csrf
        cookies = self.environ.get('HTTP_COOKIE', '') or None

        if cookies:
            cookies = dict(split_header_words([cookies])[0])

        response = self.session.get(match.group(1), cookies=cookies)
        return response.content

    def __call__(self, environ, start_response):
        self.environ = environ
        for x in self.application(environ, start_response):
            yield esi_include_re.sub(self.include, x)
